import numpy as np
import pandas as pd
from dsbox.datapreprocessing.profiler import profile_data

def isCategorical(col):
    """ 
    hardcoded rule for identifying (integer) categorical column
    """
    return col.value_counts().head(10).sum() / float(col.count()) > .95


def text2int(col):
    """
    convert column value from text to integer codes (0,1,2...)
    """
    return pd.DataFrame(col.astype('category').cat.codes,columns=[col.name])


def onehot_encode(col, exist_nan):
    """
    
    """
    return pd.get_dummies(col,prefix=col.name, dummy_na=exist_nan)

def encode(data_path):
    """

    """
    del_col = []
    new_col = []
    isDF = True

    ## csv as input ##
    if not isinstance(data_path, pd.DataFrame):
        isDF = False
        data = pd.read_csv(data_path)
        data_raw = pd.read_csv(data_path, dtype=str)
    ## data frame as input ##
    else:
        data = data_path

    for column_name in data:
        col = data[column_name].copy()
        
        exist_nan = (col.isnull().sum() > 0)

        # empty column (all missing/NaN)
        if col.count() == 0:
            print '...Delete *'+str(column_name)+"* column: empty column."
            del_col.append(col.name)

        # dtype = integer
        elif col.dtype.kind in np.typecodes['AllInteger']+'u':
            if isCategorical(col):
                print "...Delete *"+str(column_name)+"* column: integer category." 
                del_col.append(col.name)
                print "...Insert columns to onehot encode *"+str(column_name)+"*."
                new_col.append(onehot_encode(col, exist_nan))

        # dtype = float from csv file 
        # (check if it's int column with missing value)
        elif col.dtype.kind == 'f':
            if not isDF:
                col = data_raw[column_name].copy()
                pf = profile_data(pd.DataFrame(col))[column_name]['numeric_stats']
                if ('integer' in pf and 'decimal' not in pf):
                    if isCategorical(col):
                        print '...Delete *'+str(column_name)+'* column: integer category.'
                        del_col.append(col.name)
                        print "...Insert columns to onehot encode *"+str(column_name)+"*."
                        new_col.append(onehot_encode(col,exist_nan))
        
        # dtype = category
        elif col.dtype.name == 'category':
            print '...Delete *'+column_name+'* column: category dtype.'
            del_col.append(col.name)
            print "...Insert columns to onehot encode *"+str(column_name)+"*."
            new_col.append(onehot_encode(col,exist_nan))

        # for other dtypes
        else:
            #col = col.astype(str)
            if isCategorical(col):
                print '...Delete *'+str(column_name)+'* column: object/other category.'
                del_col.append(col.name)
                print "...Insert columns to onehot encode *"+str(column_name)+"*."
                new_col.append(onehot_encode(col,exist_nan))
            else:
                print '...Convert *'+str(column_name)+'* column from text to integer codes.'
                del_col.append(col.name)
                new_col.append(text2int(col))

    # drop empty, category-like and those converted to int codes
    rest = data.drop(del_col, axis=1)

    # insert columns from onehot encoding and text2int codes
    result = pd.concat(new_col+[rest], axis=1)  

    return result



import numpy as np
import pandas as pd

import missing_value_pred as mvp
import helper_func as hf


class Imputation(object):
    """search for best Imputation combination for a given dataset 

    Parameters:
    ----------
    strategies: list of string
        The Imputation strategies that are considered to be searched.

    verbose: Integer
        Control the verbosity

    eva_model: string
        The machine learning model that will be used to evaluate the imputation strategies
            support: 'linearSVM', 'decisionTree', 'logisticRegreesion', 


    Attributes:
    ----------
    best_imputation: list of string
        After 'fit' method, the best imputation combination will be given.
    """

    def __init__(self, model, scorer, greater_is_better=True, verbose=0, strategies=["mean", "max", "min", "zero"]):
        self.allowed_strategies = ["mean", "max", "min", "zero"] 
        self.verbose = verbose
        self.imputation_strategies = strategies
        self.model = model
        self.scorer = scorer


    def fit(self, data, label):
        """
        imputation combination evaluations
        now, only for classification problem (because I convert the label to integer)

        Parameters:
        ----------
        data: pandas dataframe
        label: pandas dataframe, assume second column is the label targets
        """
        
        # 1. convert categorical to indicator
        label_col_name = label.keys()[1]    # assume the second column is the label targets
        
        for col_name in data:
            if(mvp.isCategorical(data[col_name]) != None):
                data = hf.cate2ind(data, col_name)
        # convert the label also, but not to indicator, convert to integer
        data[label_col_name] = label[label_col_name] 

        if (data[label_col_name].dtypes != int and data[label_col_name].dtypes != float):
            cate_map = mvp.cate2int(data[label_col_name].unique())
            if self.verbose:
                print "convert label to integer: {}".format(cate_map)
            data[label_col_name] = data[label_col_name].replace(cate_map)

        # 2. start evaluation
        print "=========> Baseline:"
        self.__baseline(data, label_col_name)
        print "=========> Greedy searched imputation:"
        self.__imputationGreedy(data, label_col_name)
        # print "=========> other imputation method:"
        # self.__otherImpute(data, label_col_name)

        return self.best_imputation

    def impute(self, data, strategies, verbose):
        """
        impute the data using given strategies
        Parameters:
        ----------
        data: pandas dataframe
        strategies: list of string
            imputation strategies combination
        """
        for each in strategies:
            if each not in self.allowed_strategies:
                raise ValueError("Can only use these strategies: {0} "
                             " got strategy '{1}' ".format(self.allowed_strategies,
                                                        each))

        # 1. convert to np array and get missing value column id
        missing_col_id = []
        data, label = self.__df2np(data, "", missing_col_id) # no need for label
        if (len(missing_col_id) != len(strategies)):
            raise ValueError("Expected {0} number of permutations, "
                             " got '{1}' ".format(len(missing_col_id),
                                                        len(strategies)))
        # 2. impute data
        data_clean = mvp.imputeData(data, missing_col_id, strategies, verbose)

        return data_clean

    def __otherImpute(self, data, label_col_name):
        from fancyimpute import BiScaler, KNN, NuclearNormMinimization, SoftImpute
        missing_col_id = []
        data, label = self.__df2np(data, label_col_name, missing_col_id)
        # data_clean = KNN(k=5).complete(data)
        data_clean = NuclearNormMinimization().complete(data)
        self.__evaluation(data_clean, label)


    def __baseline(self, data, label_col_name):
        """
        running baseline
        """
        data_dropCol = data.dropna(axis=1, how="any") #drop the col with nan

        label = data[label_col_name].values
        data = data.drop(label_col_name,axis=1).values  #convert to np array
        label_dropCol = data_dropCol[label_col_name].values
        data_dropCol = data_dropCol.drop(label_col_name,axis=1).values

        #========================STEP 2: pred==============
        print "==============result for baseline: drop rows============"
        self.__evaluation(data, label)
        print "==============result for baseline: drop columns============"

        self.__evaluation(data_dropCol, label_dropCol)
        print "========================================================"


    def __df2np(self, data, label_col_name, missing_col_id):
        """
        helper function: convert dataframe to np array
        """
        counter = 0    
        
        # 1. get the id for missing value column
        for col_name in data:
            num = sum(pd.isnull(data[col_name]))
            if (num > 0):
                missing_col_id.append(counter)
            counter += 1

        # 2. convert the dataframe to np array
        label = None
        col_names = data.keys()
        if (len(label_col_name)>0):
            label = data[label_col_name].values
        data = data.drop(label_col_name,axis=1).values  #convert to np array

        return data, label

    def __imputationGreedy(self, data, label_col_name):
        """
        running greedy search for imputation combinations
        """        
        for each in self.imputation_strategies:
            if each not in self.allowed_strategies:
                raise ValueError("Can only use these strategies: {0} "
                             " got strategy '{1}' ".format(self.allowed_strategies,
                                                        each))
        
        col_names = data.keys()
        # 1. convert to np array and get missing value column id
        missing_col_id = []
        data, label = self.__df2np(data, label_col_name, missing_col_id)
        
        # init for the permutation
        permutations = [0] * len(missing_col_id)   # length equal with the missing_col_id; value represents the id for imputation_strategies
        pos = len(permutations) - 1
        min_score = float("inf")
        max_score = -float("inf")
        max_strategy_id = 0  
        best_combo = [0] * len(missing_col_id)  #init for best combo

        for i in range(len(permutations)):
            for strategy in range(len(self.imputation_strategies)):
                permutations[i] = strategy
                imputation_list = [self.imputation_strategies[x] for x in permutations]

                data_clean = mvp.imputeData(data, missing_col_id, imputation_list, self.verbose)
                print "for the missing value imputation combination: {} ".format(permutations)
                score = self.__evaluation(data_clean, label)
                if (score > max_score):
                    max_score = score
                    max_strategy_id = strategy
                    best_combo = permutations
                min_score = min(score, min_score)

            permutations[i] = max_strategy_id
            max_strategy_id = 0

        print "max score is {}, min score is {}\n".format(max_score, min_score)
        print "and the best score is given by the imputation combination: "
        for i in range(len(best_combo)):
            print self.imputation_strategies[best_combo[i]] + " for the column {}".format(col_names[missing_col_id[i]])

        self.best_imputation = [self.imputation_strategies[i] for i in best_combo]

    def __evaluation(self, data_clean, label):
        """
        INPUT
        data_clean: the clean dataset, missing values imputed already
        label: the label for data_clean
        """ 
        from sklearn.model_selection import train_test_split

        try:
            X_train, X_test, y_train, y_test = train_test_split(data_clean, label, test_size=0.4, random_state=0, stratify=label)
        except:
            print "cannot stratified sample, try random sample: "
            X_train, X_test, y_train, y_test = train_test_split(data_clean, label, test_size=0.4, random_state=0)
        # remove the nan rows

        mask_train = np.isnan(X_train).any(axis=1)  # nan rows index
        mask_test = np.isnan(X_test).any(axis=1)
        num_removed_test = sum(mask_test)
        X_train = X_train[~mask_train]
        y_train = y_train[~mask_train]
        X_test = X_test[~mask_test]
        y_test = y_test[~mask_test]

        model = self.model.fit(X_train, y_train)
        score = self.scorer(model, X_test, y_test)  # refer to sklearn scorer: score will be * -1 with the real score value
        print "score is: {}".format(score)

        print "===========>> max score is: {}".format(score)
        if (num_removed_test > 0):
            print "BUT !!!!!!!!there are {} data (total test size: {})that cannot be predicted!!!!!!\n".format(num_removed_test, y_test.shape[0])
        return score



import missing_value_pred as mvp
import pandas as pd
from dsbox.datapreprocessing.profiler import profile_data

def cate2ind(df, key):
    """
    missing value will also be treated as a new value
    data: pandas dataframe
    key: the key of the specified column to convert
    """
    df = df.copy()  # not modify the input
    column = df[key]
    cate_map = mvp.isCategorical(column)
    print key
    print cate_map
    if (cate_map == None):
        return df

    if (len(cate_map)==2):  # if is binary, only leave one
        cate_map.pop(cate_map.keys()[0])
    # map -> ind with 0/1
    for category in cate_map:
        new_col = column.apply(lambda x: int(x==category)) # equal => 1, not equal => 0
        if (pd.isnull(category)):    # rename the nan
            category = "missing"
        new_name = key + "/" + str(category)
        df[new_name] = new_col

    df = df.drop(key, axis=1)    # delete the original column

    return df

def miss2ind(df, key):
    """
    convert a column from given dataframe to missing/not missing (1/0), drop the original one

    """
    df = df.copy()  # not modify the input
    column = df[key]
    new_col = pd.isnull(column)
    new_name = key + "/" + "missing"
    df[new_name] = new_col
    df = df.drop(key, axis=1)

    return df

def dataPrep(data_name, label_name, drop_col_name):
    """
    data preparation:
    1. read the data into pandas dataframe
    2. drop empty and specified columns
    """
    profiler = profile_data(data_name)
    for col_name in profiler:
        if (profiler[col_name]["missing"]["num_nonblank"] == 0):
            drop_col_name.append(col_name)

    print "droped columns: {}".format(drop_col_name)
    # read data and drop
    data = pd.read_csv(data_name)   
    label = pd.read_csv(label_name)  
    for each in drop_col_name:
        data = data.drop(each, axis=1)
    return data, label


from setuptools import setup

setup(name='dsbox-datacleaning',
      version='0.1.1',
      url='https://github.com/usc-isi-i2/dsbox-cleaning.git',
      maintainer_email='kyao@isi.edu',
      maintainer='Ke-Thia Yao',
      description='DSBox data preprocessing tools for cleaning data',
      license='MIT',
      packages=['dsbox', 'dsbox.datapreprocessing', 'dsbox.datapreprocessing.cleaner'],
      zip_safe=False,
      install_requires=['pandas', 'numpy', 'scipy', 'scikit-learn', 'dsbox-dataprofiling']
      )



import pandas as pd
import numpy as np

def isCategorical(column):
    """
    INPUT
    column: pandas.dataframe series
    prematrue method: to see if 95% value is in 10 category
    if is true, will fill in the cate map that: category -> int; if not, return None
    if the cell value is nan(missing value), ignore it, leave to the imputation later
    """
    if (column.dtype == int or column.dtype == float):
        return None
    #column = column.dropna()
    total_len = len(column)
    cate = dict()
    for cell in column:
        if (len(cate) > 10):
            return None
        if (cell in cate):
            cate[cell] += 1
        else:
            cate[cell] = 1

    sorted_value = sorted(cate.values(), reverse = True)
    top10_ratio = sum(sorted_value[0:10])/float(total_len)
    if (top10_ratio >= 0.95):
        return cate2int(cate.keys())
    else:
        return None

def cate2int(cate):
    """
    return a map (dict) that maps category code to integer: 1,2,3, ... (save 0 for the new value!)
    INPUT
    cate: is a category list
    """
    category_map = {}
    ind = 1
    for i in cate:
        category_map[i] = ind
        ind += 1
    return category_map

def popular_value(array):
    """
    array: 1D array
    """
    cate = dict()
    for cell in array:
        if (cell in cate):
            cate[cell] += 1
        else:
            cate[cell] = 1

    max_freq = 0
    popular = None
    for key in cate.keys():
        if (cate[key] > max_freq):
            popular = key

    return popular

def myImputer(data, value="zero", verbose=0):
    """
    INPUT:
    data: numpy array, 1*D (one column)
    value:    string: "mean", "min", "max", "zero", "gaussian"
    """
    index = np.isnan(data)
    data_imputed = np.copy(data)
    data_drop = data[np.logical_not(index)]   #drop nan from data
    inputed_value = 0
    if (value == "zero"):
        inputed_value = 0
    elif (value == "mean"):
        inputed_value = np.mean(data_drop)
    elif (value == "max"):
        inputed_value = np.max(data_drop)
    elif (value == "min"):
        inputed_value = np.min(data_drop)
    elif (value == "new"):
        inputed_value = 0   # 0 is the value that never happens in our categorical map
    elif (value == "popular"):
        inputed_value = popular_value(data_drop)

    data_imputed[index] = inputed_value

    if verbose: print "imputed missing value: {}".format(inputed_value)
    return data_imputed


def imputeData(data, missing_col_id, imputation_strategies, verbose):
    """
    impute the data using permutations array.
    """
    data_clean = np.copy(data)

    for i in range(len(imputation_strategies)):
        strategy = imputation_strategies[i]
        col_id = missing_col_id[i]

        data_clean[:,col_id] = myImputer(data[:,col_id], strategy)

    return data_clean


#========================two learning & prediction method==============

def pred(data_clean, label, eva_model):
    """
    INPUT
    data_clean: the clean dataset, missing values imputed already
    label: the label for data_clean
    """
    from sklearn.model_selection import train_test_split
    from sklearn import svm
    from sklearn import metrics
    from sklearn.metrics import f1_score
    from sklearn.linear_model import LogisticRegressionCV
    from sklearn.linear_model import LinearRegression
    from sklearn import tree

    X_train, X_test, y_train, y_test = train_test_split(data_clean, label, test_size=0.4, random_state=0, stratify=label)
    # remove the nan rows
    mask_train = np.isnan(X_train).any(axis=1)  # nan rows index
    mask_test = np.isnan(X_test).any(axis=1)
    num_removed_test = sum(mask_test)
    X_train = X_train[~mask_train]
    y_train = y_train[~mask_train]
    X_test = X_test[~mask_test]
    y_test = y_test[~mask_test]

    # classifers:
    clf_linear = svm.SVC(kernel='linear').fit(X_train, y_train)
    clf_lg = LogisticRegressionCV(scoring="f1_macro").fit(X_train, y_train)
    clf_tree = tree.DecisionTreeClassifier().fit(X_train, y_train)

    max_score = 0

    if (eva_model=="linearSVM"):
        def scorer (y_true, y_pred):
            return f1_score(y_true, y_pred, average="macro")

        print scorer(y_test, clf_linear.predict(X_test))
        # print "for linear SVM: "
        # print f1_score(y_test, clf_linear.predict(X_test), average=None)
        # score = f1_score(y_test, clf_linear.predict(X_test), average="macro")    #weighted average over all the classes
        # max_score = max(score, max_score)
        # print  score


    if (eva_model=="logisticRegreesion"):
        print "for LogisticRegressionCV: "
        print f1_score(y_test, clf_lg.predict(X_test), average=None)
        score = f1_score(y_test, clf_lg.predict(X_test), average="macro")   #weighted average over all the classes
        max_score = max(score, max_score)
        print score 
    
    if (eva_model=="decisionTree"):
        print "for decision tree: "
        print f1_score(y_test, clf_tree.predict(X_test), average=None)
        score = f1_score(y_test, clf_tree.predict(X_test), average="macro")   #weighted average over all the classes
        max_score = max(score, max_score)
        print score 

    print "===========>> max score is: {}".format(max_score)
    if (num_removed_test > 0):
        print "BUT !!!!!!!!there are {} data (total test size: {})that cannot be predicted!!!!!!\n".format(num_removed_test, y_test.shape[0])
    return max_score

def pred_cv(data_clean, label, eva_metrics):
    """
    prediction with cross validation.
    INPUT
    data_clean: the clean dataset, missing values imputed already
    label: the label for data_clean
    """
    from sklearn import svm
    from sklearn import metrics
    from sklearn.linear_model import LogisticRegressionCV
    from sklearn.linear_model import LinearRegression
    from sklearn import tree
    from sklearn.model_selection import cross_val_score
    from sklearn.ensemble import RandomForestClassifier

    clf_linear = svm.SVC(kernel='linear')
    clf_lg = LogisticRegressionCV(scoring=eva_metrics)
    clf_tree = tree.DecisionTreeClassifier()
    max_score = 0

    score = cross_val_score(clf_linear, data_clean, label, scoring=eva_metrics).mean()
    print "for SVM linear, score is {}".format(score)
    max_score = max(max_score, score)

    score = cross_val_score(clf_lg, data_clean, label, scoring=eva_metrics).mean()
    print "for LogisticRegression, score is {}".format(score)
    max_score = max(max_score, score)

    score = cross_val_score(clf_tree, data_clean, label, scoring=eva_metrics).mean()
    print "for DecisionTree, score is {}".format(score)
    max_score = max(max_score, score)

    print "===========>> max weighted score is: {}".format(max_score)
    return max_score
